# pylint: disable = missing-docstring

from typing import Any, List
import pytest
from typed_descriptors import Attr, Prop,cached_property

@pytest.mark.parametrize("decorator", [False, True])
@pytest.mark.parametrize("x", [0, 10])
def test_single_prop_valid(x: int, decorator: bool) -> None:
    class C:
        if not decorator:
            @cached_property
            def x(self) -> int:
                return x
        else:
            x = Prop(int, lambda _: x) # type: ignore
    c = C()
    assert c.x == x

@pytest.mark.parametrize("decorator", [False, True])
@pytest.mark.parametrize("x", [0, 10])
def test_single_prop_cache(x: int, decorator: bool) -> None:
    class C:
        if decorator:
            @cached_property
            def x(self) -> int:
                return x
        else:
            x = Prop(int, lambda _: x) # type: ignore
    c = C()
    assert not C.x.is_cached_on(c)
    assert c.x == x
    assert C.x.is_cached_on(c)

@pytest.mark.parametrize("decorator", [False, True])
@pytest.mark.parametrize("x", [0, 10])
def test_single_prop_cache_delete(x: int, decorator: bool) -> None:
    class C:
        if decorator:
            @cached_property
            def x(self) -> int:
                return x
        else:
            x = Prop(int, lambda _: x) # type: ignore
    c = C()
    assert c.x == x
    del c.x
    assert not C.x.is_cached_on(c)
    assert c.x == x
    assert C.x.is_cached_on(c)

@pytest.mark.parametrize("decorator", [False, True])
@pytest.mark.parametrize("x", ["a", [0, 1]])
def test_single_prop_type_error(x: Any, decorator: bool) -> None:
    class C:
        if decorator:
            @cached_property
            def x(self) -> int:
                return x # type: ignore
        else:
            x = Prop(int, lambda _: x) # type: ignore
    c = C()
    with pytest.raises(TypeError):
        assert c.x == x

@pytest.mark.parametrize("decorator", [False, True])
@pytest.mark.parametrize("y", [[], [0, 1, 2]])
def test_prop_attr_dependency(y: List[int], decorator: bool) -> None:
    class C:
        if decorator:
            @cached_property
            def x(self) -> int:
                return len(self.y)
        else:
            x = Prop(int, lambda self: len(self.y)) # type: ignore
        y = Attr(List[int])
    c = C()
    c.y = y
    assert c.x == len(y)

@pytest.mark.parametrize("decorator", [False, True])
@pytest.mark.parametrize("y", [[], [0, 1, 2]])
def test_prop_attr_type_error(y: List[int], decorator: bool) -> None:
    class C:
        if decorator:
            @cached_property
            def x(self) -> int:
                return self.y # type: ignore
        else:
            x = Prop(int, lambda self: self.y) # type: ignore
        y = Attr(List[int])
    c = C()
    c.y = y
    with pytest.raises(TypeError):
        assert c.x == len(y)



# pylint: disable = missing-docstring

import sys
from typing import Union
import pytest
from typed_descriptors import Attr

if sys.version_info[1] >= 9:
    from collections.abc import Sequence
else:
    from typing import Sequence

def test_getting_started_example() -> None:
    # pylint: disable = import-outside-toplevel

        class MyClass:

            x = Attr(int, lambda _, x: x >= 0, readonly=True)
            y = Attr(Sequence[int], lambda self, y: len(y) <= self.x)

            def __init__(self, x: int, y: Sequence[int]):
                self.x = x
                self.y = y

        myobj = MyClass(3, [0, 2, 5]) # OK
        myobj.y = (0, 1)              # OK
        with pytest.raises(ValueError):
            myobj.y = [0, 2, 4, 6]
        with pytest.raises(AttributeError):
            myobj.x = 5
        with pytest.raises(TypeError):
            myobj.y = 5
        with pytest.raises(TypeError):
            myobj.y = ["hi", "bye"]

def test_attr_validator_example_1() -> None:
    class MyClass:
        @Attr.validator
        def x(self, value: Sequence[str]) -> bool:
            ''' Validator function for attribute 'C.x'. '''
            return 1 <= len(value) <= 3

    myobj = MyClass()
    myobj.x = ["a", "b", "c"]
    assert myobj.x == ["a", "b", "c"]
    assert "x" in myobj.__dict__
    assert myobj.__dict__["x"] == ["a", "b", "c"]
    with pytest.raises(TypeError):
        myobj.x = 10
    with pytest.raises(ValueError):
        myobj.x = ["a", "b", "c", "d"]

def test_attr_validator_example_2() -> None:
    class MyClass:
        @Attr.validator(readonly=True)
        def x(self, value: Union[int, str]) -> bool:
            ''' Validator function for readonly attribute 'C.x'. '''
            if isinstance(value, int):
                return value > 0
            return len(value) > 0

        @Attr.validator(backed_by='_w')
        def w(self, value: Sequence[int]) -> bool:
            ''' Validator function for mutable attribute 'C.w'. '''
            return len(value) in range(3)

        __slots__ = ("__x", "_w")
        #    default ^^^^^  ^^^^ custom
        #        backing attributes

    myobj = MyClass()
    myobj.x = "hello"
    assert myobj.x == "hello"
    assert hasattr(myobj, "_MyClass__x")
    assert getattr(myobj, "_MyClass__x") == "hello"
    with pytest.raises(AttributeError):
        myobj.x = "hello"
    with pytest.raises(AttributeError):
        myobj.x = 1.5
    with pytest.raises(AttributeError):
        myobj.x = ""



"""
    Descriptor class for attributes.
"""

# Part of typed-descriptors
# Copyright (C) 2023 Hashberg Ltd

from __future__ import annotations
from inspect import signature
from typing import (
    Any,
    Optional,
    Protocol,
    Type,
    TypeVar,
    Union,
    final,
    get_type_hints,
    overload,
)
from typing_extensions import Self
from typing_validation import validate
from .base import DescriptorBase, T


_T = TypeVar("_T")
""" Invariant type variable for generic values, privately used. """

T_contra = TypeVar("T_contra", contravariant=True)
""" Contravariant type variable for generic values. """


class SupportsBool(Protocol):
    """
    Structural types for things which can be converted to :obj:`bool`.
    """

    def __bool__(self) -> bool: ...


class ValidatorFunction(Protocol[T_contra]):
    """
    Structural type for the validator function of a :class:`Attr`.
    """

    def __call__(
        self, instance: Any, value: T_contra, /
    ) -> Union[SupportsBool, None]:
        """
        Validates the given value for assignment to a :class:`Attr`,
        in the context of the given instance.

        Called passing the current ``instance`` and the ``value`` that is
        to be assigned to the descriptor.
        At the time when the validator function for a descriptor is invoked,
        the given ``value`` has already passed its runtime typecheck.
        Validator functions can use ``instance`` to perform validation
        involving other descriptors for the same class.

        There are two ways in which a validator function can trigger an error
        as part of the validation logic for an :class:`Attr`:

        - By returning :obj:`False`: a :obj:`ValueError` will be raised.
        - By raising any exception as part of its body: the exception is
          caught as part of a `try...except` block, and a :obj:`ValueError`
          is raised from it (preserving the original error information).

        In the second case, the validator should return :obj:`True` at the
        end, to signal that validation was successful.

        """
        ...


def validate_validator_fun(validator_fun: ValidatorFunction[T], /) -> None:
    """
    Runtime validation for validator functions.

    :raises TypeError: if the argument is not a validator function
    """
    if not callable(validator_fun):
        raise TypeError("Validator function must be callable.")
    if len(signature(validator_fun).parameters) != 2:
        raise TypeError("Validator function must take exactly two arguments.")


def validator_fun_value_type(validator_fun: ValidatorFunction[T], /) -> Any:
    """
    Returns the type annotation for the ``value`` argument of a validator
    function. Used by :meth:`Attr.validator` to infer the :class:`Attr`
    type from the validator function type hints.

    :raises TypeError: if the argument is not a validator function
    :raises ValueError: if the function doesn't have an explicit annotation
                        for its ``value`` argument type or its return type.
    """
    validate_validator_fun(validator_fun)
    validator_fun_types = get_type_hints(validator_fun)
    validator_fun_sig = signature(validator_fun)
    validator_fun_params = tuple(validator_fun_sig.parameters.keys())
    value_argname = validator_fun_params[1]
    if value_argname not in validator_fun_types:
        raise ValueError(
            "Validator function must explicitly annotate the type for its "
            f"second argument {value_argname!r}."
        )
    return validator_fun_types[value_argname]


class ValidatedAttrFactory(Protocol):
    """
    Structural type for functions which create :class:`Attr` instances
    from validator functions.
    """

    def __call__(self, validator_fun: ValidatorFunction[T], /) -> Attr[T]:
        """
        Returns a validated :class:`Attr` from a validator function.
        """
        ...


class Attr(DescriptorBase[T]):
    """
    A descriptor class for attributes, supporting:

    - static type checking for the attribute value
    - runtime type checking of values assigned to the attribute
    - optional runtime validation of values assigned to the attribute
    - optional read-only restrictions on the attribute (set once)

    See :class:`DescriptorBase` for details on how the attribute value is
    stored in each instance.
    """

    @staticmethod
    @overload
    def validator(
        validator_fun: ValidatorFunction[T],
        /,
        *,
        readonly: bool = False,
        backed_by: Optional[str] = None,
        use_dict: Optional[bool] = None,
    ) -> Attr[T]: ...

    @staticmethod
    @overload
    def validator(
        validator_fun: None = None,
        /,
        *,
        readonly: bool = False,
        backed_by: Optional[str] = None,
        use_dict: Optional[bool] = None,
    ) -> ValidatedAttrFactory: ...

    @staticmethod
    def validator(
        validator_fun: Optional[ValidatorFunction[T]] = None,
        /,
        *,
        readonly: bool = False,
        backed_by: Optional[str] = None,
        use_dict: Optional[bool] = None,
    ) -> ValidatedAttrFactory | Attr[T]:
        """
        Decorator used to create an :class:`Attr` from a validator function,
        optionally specifying a readonly modifier and a backing attribute.

        It can be used directly, for mutable attributes with default backing
        attribute name:

            .. code-block ::

                class MyClass:

                    @Attr.validator
                    def x(self, value: Sequence[str]) -> bool:
                        ''' Validator function for attribute 'C.x'. '''
                        return 1 <= len(value) <= 3

        It can be used by supplying ``readonly`` and ``backed_by`` values,
        for more general attributes:

            .. code-block ::

                class MyClass:

                    @Attr.validator(readonly=True)
                    def x(self, value: int|str) -> bool:
                        ''' Validator function for readonly attribute 'C.x'. '''
                        if isinstance(value, int):
                            return value > 0
                        return len(value) > 0

                    @Attr.validator(backed_by='_w')
                    def w(self, value: Sequence[int]) -> bool:
                        ''' Validator function for mutable attribute 'C.w'. '''
                        return len(value) in range(3)

                    __slots__ = ("__x", "_w")
                    #    default ^^^^^  ^^^^ custom
                    #        backing attributes

        """
        if validator_fun is not None:
            ty = validator_fun_value_type(validator_fun)
            return Attr(
                ty,
                validator=validator_fun,
                readonly=readonly,
                backed_by=backed_by,
                use_dict=use_dict,
            )

        def _validated_attr(validator_fun: ValidatorFunction[_T]) -> Attr[_T]:
            return Attr.validator(
                validator_fun,
                readonly=readonly,
                backed_by=backed_by,
                use_dict=use_dict,
            )

        return _validated_attr

    __readonly: bool
    __validator_fun: Optional[ValidatorFunction[T]]

    @overload
    def __init__(
        self,
        type: Type[T],
        /,
        validator: Optional[ValidatorFunction[T]] = None,
        *,
        readonly: bool = False,
        backed_by: Optional[str] = None,
        use_dict: Optional[bool] = None,
    ) -> None:
        # pylint: disable = redefined-builtin
        ...

    @overload
    def __init__(
        self,
        type: Any,
        /,
        validator: Optional[ValidatorFunction[T]] = None,
        *,
        readonly: bool = False,
        backed_by: Optional[str] = None,
        use_dict: Optional[bool] = None,
    ) -> None:
        # pylint: disable = redefined-builtin
        ...

    def __init__(
        self,
        type: Type[T] | Any,
        /,
        validator: Optional[ValidatorFunction[T]] = None,
        *,
        readonly: bool = False,
        backed_by: Optional[str] = None,
        use_dict: Optional[bool] = None,
    ) -> None:
        """
        Creates a new attribute with the given type and optional validator.

        :param ty: the type of the attribute
        :param validator: an optional validator function for the attribute
        :param readonly: whether the attribute is read-only

        :raises TypeError: if the type is not a valid type
        :raises TypeError: if the validator is not callable

        :meta public:
        """
        # pylint: disable = redefined-builtin
        super().__init__(type, backed_by=backed_by, use_dict=use_dict)
        if validator is not None:
            validate_validator_fun(validator)
            self.__doc__ = validator.__doc__
        self.__validator_fun = validator
        self.__readonly = bool(readonly)

    @final
    @property
    def readonly(self) -> bool:
        """
        Whether the attribute is readonly.
        """
        return self.__readonly

    @final
    @property
    def validator_fun(self) -> Optional[ValidatorFunction[T]]:
        """
        The custom validator function for the attribute,
        or :obj:`None` if no validator was specified.

        There are two ways in which the validator can trigger an error:

        - By returning :obj:`False`: a :obj:`ValueError` is raised
        - By raising any exception as part of its body: the exception is
          caught as part of a `try...except` block, and a :obj:`ValueError`
          is raised from it (preserving the original error information).

        In the second case, the validator should return :obj:`True` at the
        end, to signal that validation was successful.
        """
        return self.__validator_fun

    @final
    def is_defined_on(self, instance: Any) -> bool:
        """
        Wether the attribute is defined on the given instance.
        """
        validate(instance, self.owner)
        return self._is_set_on(instance)

    @overload
    def __get__(self, instance: None, _: Type[Any]) -> Self: ...

    @overload
    def __get__(self, instance: Any, _: Type[Any]) -> T: ...

    @final
    def __get__(self, instance: Any, _: Type[Any]) -> T | Self:
        """
        If the descriptor is accessed on an instance, returns the value of
        the attribute on the given instance.

        If the descriptor is accessed on the owner class, i.e. if
        ``instance`` is :obj:`None`, returns the :class:`Attr` object.

        :raises AttributeError: if the attribute is not defined,
                                see :meth:`is_defined_on`.

        :meta public:
        """
        if instance is None:
            return self
        try:
            return self._get_on(instance)
            # return cast(T, getattr(instance, self.attr_name))
        except AttributeError:
            raise AttributeError(f"Attribute {self} is not set.") from None

    @final
    def __set__(self, instance: Any, value: T) -> None:
        """
        Sets the value of the descriptor on the given instance.

        :raises AttributeError: if the attribute is readonly and it already
                                has a value assigned to it
        :raises TypeError: if the value has the wrong type
        :raises ValueError: if a custom validator is specified and the
                            value is invalid

        :meta public:
        """
        if self.readonly:
            if self._is_set_on(instance):
                raise AttributeError(
                    f"Attribute {self} is readonly: it can only be set once."
                )
        validate(value, self.type)
        validator = self.__validator_fun
        if validator is not None:
            try:
                res = validator(instance, value)
                if res is not None and not res:
                    raise ValueError(
                        f"Invalid value for attribute {self}: {value!r}"
                    )
            except ValueError as e:
                raise ValueError(
                    f"Invalid value for attribute {self}: {value!r}"
                ) from e
        self._set_on(instance, value)

    @final
    def __delete__(self, instance: Any) -> None:
        """
        Deletes the value of the descriptor on the given instance.

        :raises AttributeError: if the attribute is readonly
        :raises AttributeError: if the attribute is not defined,
                                see :meth:`is_defined_on`.

        :meta public:
        """
        if self.readonly:
            raise AttributeError(
                f"Attribute {self.name!r} is readonly: it cannot be deleted."
            )
        if not self._is_set_on(instance):
            raise AttributeError(f"Attribute {self} is not set.")
        self._del_on(instance)

    def __str__(self) -> str:
        """
        Representation of this attribute, inclusive of the following info:

        - the :attr:`owner` name
        - the attribute :attr:`name`

        An example:

        .. code-block ::

            Color.hue
        """
        owner_name = self.owner.__name__
        name = self.name
        return f"{owner_name}.{name}"

    def __repr__(self) -> str:
        """
        Representation of this attribute, inclusive of the following info:

        - the :attr:`owner` name
        - the attribute :attr:`name`
        - the attribute :attr:`type`
        - optional ``readonly`` qualifier

        Two examples:

        .. code-block ::

            <Attr Color.hue: str>
            <readonly Attr Color.saturation: int>

        """
        owner = self.owner.__name__
        name = self.name
        ty = (
            self.type.__name__
            if isinstance(self.type, type)
            else str(self.type)
        )
        qualifier = "readonly " if self.readonly else ""
        return f"<{qualifier}Attr {owner}.{name}: {ty}>"



"""
    Descriptor classes with static and runtime validation features.
"""

# typed-descriptors: Descriptor classes for typed attributes and properties.
# Copyright (C) 2023 Hashberg Ltd

from __future__ import annotations
from .base import TypedDescriptor
from .attr import Attr
from .prop import Prop, cached_property

__version__ = "1.2.1"

__all__ = ("TypedDescriptor", "Attr", "Prop", "cached_property")



"""
A Mypy plugin to enhance static type inference of descriptors from
the :mod`typed_descriptors` library.
"""

from __future__ import annotations
from collections.abc import Callable
import typing
from mypy.plugin import FunctionContext, Plugin
from mypy.types import CallableType, Instance, get_proper_type, Type


def typed_descriptor_hook(ctx: FunctionContext) -> Type:
    """
    Extracts the descriptor type and assigns it to the generic type parameter,
    presuming that the generic type parameter appears at index 0.
    """
    assert ctx.arg_types and ctx.arg_types[0]
    _descriptor_t = get_proper_type(ctx.arg_types[0][0])
    ret_t = get_proper_type(ctx.default_return_type)
    if isinstance(_descriptor_t, CallableType):
        descriptor_t = _descriptor_t.ret_type
        if isinstance(ret_t, Instance):
            args = list(ret_t.args)
            args[0] = descriptor_t
            return ret_t.copy_modified(args=tuple(args))
        return ret_t
    return ret_t


_function_hooks = {
    "typed_descriptors.attr.Attr": typed_descriptor_hook,
    "typed_descriptors.prop.Prop": typed_descriptor_hook,
}


class TypedDescriptorsPlugin(Plugin):
    """
    Mypy plugin which expands type inference for typed descriptor,
    in the absence of TypeForm.
    """

    def get_function_hook(
        self, fullname: str
    ) -> Callable[[FunctionContext], Type] | None:
        hook = _function_hooks.get(fullname)
        if hook is not None:
            return hook
        return super().get_function_hook(fullname)


def plugin(version: str) -> typing.Type[TypedDescriptorsPlugin]:
    """
    Entry point for the Mypy plugin.
    """
    # ignore version argument if the plugin works with all mypy versions.
    return TypedDescriptorsPlugin



# pylint: disable = missing-docstring

import sys
from typing import Optional
import pytest
from typed_descriptors import Attr
from typed_descriptors.base import name_mangle

@pytest.mark.parametrize("backed_by", [None, "x", "_x", "_y", "__x", "__y"])
def test_base_set_name(backed_by: Optional[str]) -> None:
    if backed_by is None:
        backing_attr = "x"
    else:
        backing_attr = backed_by
    class C:
        x = Attr(int, lambda _, x: x >= 0, backed_by=backed_by)
        def __init__(self, x: int) -> None:
            self.x = x
    c = C(0)
    assert name_mangle(C, backing_attr) in c.__dict__

@pytest.mark.parametrize("backed_by", [None, "_x", "_y", "__x", "__y"])
def test_base_set_name_slots(backed_by: Optional[str]) -> None:
    if backed_by is None:
        backing_attr = "__x"
    else:
        backing_attr = backed_by
    class C:
        x = Attr(int, lambda _, x: x >= 0, backed_by=backed_by)
        __slots__ = (backing_attr,)
        def __init__(self, x: int) -> None:
            self.x = x
    c = C(0)
    assert hasattr(c, name_mangle(C, backing_attr))

@pytest.mark.parametrize("backed_by", [None, "x", "_x", "_y", "__x", "__y"])
def test_base_set_name_slots_dict(backed_by: Optional[str]) -> None:
    if backed_by is None:
        backing_attr = "x"
    else:
        backing_attr = backed_by
    class C:
        x = Attr(int, lambda _, x: x >= 0, backed_by=backed_by)
        __slots__ = ("__dict__",)
        def __init__(self, x: int) -> None:
            self.x = x
    c = C(0)
    assert name_mangle(C, backing_attr) in c.__dict__

name_slot_error = (
    TypeError
    if sys.version_info[1] >= 12
    else RuntimeError
)

@pytest.mark.parametrize("backed_by", [None, "x", "_x", "_y", "__x", "__y"])
def test_base_set_name_slots_error(backed_by: Optional[str]) -> None:
    with pytest.raises(name_slot_error):
        class C:
            x = Attr(int, lambda _, x: x >= 0, backed_by=backed_by)
            __slots__ = ("some_other_attr")
            # Recall that __slots__ = () is interpreted as not declaring slots



"""
    Descriptor class for cached properties.
"""

# Part of typed-descriptors
# Copyright (C) 2023 Hashberg Ltd

from __future__ import annotations
from inspect import signature
import sys
from typing import (
    Any,
    Optional,
    Protocol,
    Type,
    TypeVar,
    final,
    get_type_hints,
    overload,
)
from typing_extensions import Self
from typing_validation import validate
from .base import DescriptorBase, T


_T = TypeVar("_T")
""" Invariant type variable for generic values, privately used. """

T_co = TypeVar("T_co", covariant=True)
""" Covariant type variable for generic values. """


class ValueFunction(Protocol[T_co]):
    """
    Structural type for the value function of a :class:`Prop`.
    """

    def __call__(self, instance: Any, /) -> T_co:
        """
        Computes and returns the value for a :class:`Prop`,
        in the context of the given instance.
        """
        ...


def validate_value_fun(value_fun: ValueFunction[T], /) -> None:
    """
    Runtime validation for value functions.

    :raises TypeError: if the argument is not a value function
    :raises ValueError: if the function doesn't have an explicit annotation
                        for its return type.
    """
    if not callable(value_fun):
        raise TypeError("Value function must be callable.")
    if len(signature(value_fun).parameters) != 1:
        raise TypeError("Value function must take exactly one argument.")


def value_fun_return_type(value_fun: ValueFunction[T], /) -> Any:
    """
    Returns the return type annotation of a value function.
    Used by :func:`Prop` to infer the property type from the value function
    return type annotation.

    :raises TypeError: if the argument is not a value function
    :raises ValueError: if the function doesn't have an explicit annotation
                        for its return type.
    """
    validate_value_fun(value_fun)
    value_fun_types = get_type_hints(value_fun)
    if "return" not in value_fun_types:
        raise ValueError(
            "Value function must explicitly annotate its return type."
        )
    return value_fun_types["return"]


class PropFactory(Protocol):
    """
    Structural type for functions which create :class:`Prop` instances
    from value functions.
    """

    def __call__(self, value_fun: ValueFunction[T], /) -> Prop[T]:
        """
        Returns a validated :class:`Prop` from a value function.
        """
        ...


class Prop(DescriptorBase[T]):
    """
    A descriptor class for cached properties, supporting:

    - static type checking for the property value
    - lazy caching (value only cached at first read)

    See :class:`~typed_descriptors.base.DescriptorBase` for details on how the property value is
    cached in each instance.
    """

    @staticmethod
    @overload
    def value(
        value_fun: ValueFunction[T],
        /,
        *,
        backed_by: Optional[str] = None,
        use_dict: Optional[bool] = None,
    ) -> Prop[T]: ...

    @staticmethod
    @overload
    def value(
        value_fun: None = None,
        /,
        *,
        backed_by: Optional[str] = None,
        use_dict: Optional[bool] = None,
    ) -> PropFactory: ...

    @staticmethod
    def value(
        value_fun: Optional[ValueFunction[T]] = None,
        /,
        *,
        backed_by: Optional[str] = None,
        use_dict: Optional[bool] = None,
    ) -> PropFactory | Prop[T]:
        """
        An alias for :func:`cached_property`.

        It offers a way to declare :class:`Prop` which is stylisticallly
        aligned to the :meth:`Attr.validator<typed_descriptors.attr.Attr.validator>`
        decorator for attributes.
        """
        return cached_property(
            value_fun, backed_by=backed_by, use_dict=use_dict
        )

    __value_fun: ValueFunction[T]

    @overload
    def __init__(
        self,
        type: Type[T],
        value: ValueFunction[T],
        /,
        *,
        backed_by: Optional[str] = None,
        use_dict: Optional[bool] = None,
    ) -> None:
        # pylint: disable = redefined-builtin
        ...

    @overload
    def __init__(
        self,
        type: Any,
        value: ValueFunction[T],
        /,
        *,
        backed_by: Optional[str] = None,
        use_dict: Optional[bool] = None,
    ) -> None:
        # pylint: disable = redefined-builtin
        ...

    def __init__(
        self,
        type: Type[T] | Any,
        value: ValueFunction[T],
        /,
        *,
        backed_by: Optional[str] = None,
        use_dict: Optional[bool] = None,
    ) -> None:
        """
        Creates a new property with the given type and value function.

        :param value: function computing the property value
        :param type: the type of the property
        :param attr_name: the name of the backing attribute for the property
                          cache, or :obj:`None` to use a default name

        :raises TypeError: if the type is not a valid type
        :raises TypeError: if the value function is not callable

        :meta public:
        """
        # pylint: disable = redefined-builtin
        validate_value_fun(value)
        super().__init__(type, backed_by=backed_by, use_dict=use_dict)
        if not callable(value):
            raise TypeError(f"Expected callable 'value', got {value!r}.")
        self.__value_fun = value
        self.__doc__ = value.__doc__

    @final
    @property
    def value_fun(self) -> ValueFunction[T]:
        """
        The function called to produce a value for this property on a given
        instance.
        It is called by the getter when the property doesn't have a cached
        value, and the value returned is automatically cached.
        """
        return self.__value_fun

    @final
    def is_cached_on(self, instance: Any) -> bool:
        """
        Whether the property is cached on the given instance.
        """
        validate(instance, self.owner)
        return self._is_set_on(instance)

    @final
    def cache_on(self, instance: Any) -> None:
        """
        Caches the property value on the given instance.
        Can be used to force property computation at a desired time,
        overriding the default lazy behaviour.

        :raises AttributeError: if the property is already cached.
        """
        validate(instance, self.owner)
        if self._is_set_on(instance):
            raise AttributeError(f"Property {self} is already cached.")
        value = self.value_fun(instance)
        validate(value, self.type)
        self._set_on(instance, value)

    @overload
    def __get__(self, instance: None, _: Type[Any]) -> Self: ...

    @overload
    def __get__(self, instance: Any, _: Type[Any]) -> T: ...

    @final
    def __get__(self, instance: Any, _: Type[Any]) -> T | Self:
        """
        If the descriptor is accessed on an instance, returns the value of
        the property on the given instance.

        If the descriptor is accessed on the owner class, i.e. if
        ``instance`` is :obj:`None`, returns the :class:`Prop` object.

        :meta public:
        """
        if instance is None:
            return self
        try:
            return self._get_on(instance)
        except AttributeError:
            value = self.value_fun(instance)
            validate(value, self.type)
            self._set_on(instance, value)
            return value

    __set__ = None
    """
        Property values cannot be set.

        :meta public:
    """

    @final
    def __delete__(self, instance: Any) -> None:
        """
        Deletes the property cache on the given instance.

        :raises AttributeError: if the property is not cached,
                                see :meth:`is_cached_on`.

        :meta public:
        """
        if not self._is_set_on(instance):
            raise AttributeError(f"Property {self} is not cached.")
        self._del_on(instance)

    def __str__(self) -> str:
        """
        Representation of this property, inclusive of the following info:

        - the :attr:`owner` name
        - the property :attr:`name`

        An example:

        .. code-block ::

            Color.hue
        """
        owner_name = self.owner.__name__
        name = self.name
        return f"{owner_name}.{name}"

    def __repr__(self) -> str:
        """
        Representation of this property, inclusive of the following info:

        - the :attr:`owner` name
        - the property :attr:`name`
        - the property :attr:`type`

        An example:

        .. code-block ::

            <Prop Color.rgb: tuple[int, int, int]>

        """
        owner = self.owner.__name__
        name = self.name
        ty = (
            self.type.__name__
            if isinstance(self.type, type)
            else str(self.type)
        )
        return f"<Prop {owner}.{name}: {ty}>"


@overload
def cached_property(
    value_fun: ValueFunction[T],
    /,
    *,
    backed_by: Optional[str] = None,
    use_dict: Optional[bool] = None,
) -> Prop[T]: ...


@overload
def cached_property(
    value_fun: None = None,
    /,
    *,
    backed_by: Optional[str] = None,
    use_dict: Optional[bool] = None,
) -> PropFactory: ...


def cached_property(
    value_fun: Optional[ValueFunction[T]] = None,
    /,
    *,
    backed_by: Optional[str] = None,
    use_dict: Optional[bool] = None,
) -> PropFactory | Prop[T]:
    """
    Decorator used to create a cached property from a value function,
    optionally specifying a backing attribute.
    See :class:`Prop` and :class:`DescriptorBase` for more information.

    It can be used directly, for properties with default backing attribute name:

    .. code-block ::

        class C:

            @cached_property
            def x(self) -> Sequence[str]:
                ''' Value function for property 'C.x'. '''
                return 10

    It can be used by supplying a custom backing attribute name to the
    ``backed_by`` argument:

    .. code-block ::

        class C:

            @cached_property(backed_by="_x")
            def x(self) -> Sequence[str]:
                ''' Value function for property 'C.x'. '''
                return 10

            __slots__ = ("_x", )

    .. note ::

        The decorator is analogous to the built-in :func:`functools.cached_property`,
        from which it takes its name, and it uses the same caching logic when
        ``__dict__`` is available on owner class's instances and no custom
        attribute name is used.
        Contrary to its built-in counterpart, however, this decorator can be
        used with a slotted attribute as backing attribute.

    """
    if value_fun is not None:
        prop_type = value_fun_return_type(value_fun)
        return Prop(
            prop_type, value_fun, backed_by=backed_by, use_dict=use_dict
        )

    def _cached_property(value_fun: ValueFunction[_T]) -> Prop[_T]:
        return cached_property(
            value_fun, backed_by=backed_by, use_dict=use_dict
        )

    return _cached_property



"""
    Abstract base class for descriptors backed by protected/private attributes.
"""

# Part of typed-descriptors
# Copyright (C) 2023 Hashberg Ltd

from __future__ import annotations
from abc import abstractmethod
import sys
from typing import (
    Any,
    Optional,
    Protocol,
    Type,
    TypeVar,
    Union,
    cast,
    final,
    overload,
    runtime_checkable,
)
from typing_extensions import Self
from typing_validation import can_validate, validate


def is_dict_available(owner: type) -> bool:
    """
    Checks whether instances of a descriptor owner class have ``__dict__``.
    Returns :obj:`True` if the MRO root ``owner.__mro__[-1]`` is not
    :obj:`object`, or if the following is true for any class ``cls`` in
    ``owner.__mro__[:-1]`` (i.e. excluding the MRO root):

    1. ``cls`` has no ``__slots__`` attribute
    2. the ``__slots__`` attribute of ``cls`` is empty
    3. ``__dict__`` appears in the the ``__slots__`` attribute of ``cls``

    Otherwise, returns :obj:`False`.

    .. warning ::

        If the function returns :obj:`False`, then ``__dict__`` is certainly
        not available. However, it is possible for a class ``cls`` in the MRO to
        satisfy one of the three conditions above and for ``__dict__`` to not be
        available on instances of the descriptor owner class. For example:

        1. The ``__slots__`` attribute might have been deleted at some point
           after the slots creation process.
        2. The ``__slots__`` attribute contents might have been cleared, or it
           might have been an iterable which as been fully consumed as part of
           the slots creation process.
        3. A ``__dict__`` entry might have been added to ``__slots__`` at some
           point after the slots creation process.

        If this happens somewhere in the MRO of the owner class being inspected,
        the library will incorrectly infer that ``__dict__`` is available on
        class instances, resulting in incorrect behaviour.
        In such situations, please manually set ``use_dict`` to :obj:`False` in
        attribute constructors to ensure that the ``__dict__`` mechanism is not
        used to back the descriptor.

    """
    mro = owner.__mro__
    if mro[-1] != object:
        return False
    for cls in mro[:-1]:
        if not hasattr(cls, "__slots__"):
            return True
        if not cls.__slots__:
            return True
        if "__dict__" in cls.__slots__:
            return True
    return False


def class_slots(cls: type) -> tuple[str, ...] | None:
    """
    Returns a tuple consisting of all slots for the given class and all
    non-private slots for all classes in its MRO.
    Returns :obj:`None` if slots are not defined for the class.
    """
    if not hasattr(cls, "__slots__"):
        return None
    slots: list[str] = list(cls.__slots__)
    for cls in cls.__mro__[1:-1]:
        for slot in getattr(cls, "__slots__", ()):
            assert isinstance(slot, str)
            if slot.startswith("__") and not slot.endswith("__"):
                continue
            slots.append(slot)
    return tuple(slots)


def name_mangle(owner: type, attr_name: str) -> str:
    """
    If the given attribute name is private and not dunder,
    return its name-mangled version for the given owner class.
    """
    if not attr_name.startswith("__"):
        return attr_name
    if attr_name.endswith("__"):
        return attr_name
    return f"_{owner.__name__}{attr_name}"


def name_unmangle(owner: type, attr_name: str) -> str:
    """
    If the given attribute name is name-mangled for the given owner class,
    removes the name-mangling prefix.
    """
    name_mangling_prefix = f"_{owner.__name__}"
    if attr_name.startswith(name_mangling_prefix + "__"):
        return attr_name[len(name_mangling_prefix) :]
    return attr_name


T = TypeVar("T")
""" Invariant type variable for generic values. """

T_co = TypeVar("T_co", covariant=True)
""" Invariant type variable for generic values. """


@runtime_checkable
class TypedDescriptor(Protocol[T_co]):
    """
    Structural type for typed descriptors.
    """

    __descriptor_type__: Any  # Will be TypeForm[T_co]
    """
    Indicates the type (or type annotation, if a string) for the descriptor.

    :meta public:
    """

    def __set_name__(self, owner: Type[Any], name: str) -> None:
        """
        Hook called when the descriptor is assigned to a class attribute,
        usually responsible for setting the owner and name of the descriptor.

        :meta public:
        """

    @overload
    def __get__(self, instance: None, _: Type[Any]) -> Self: ...

    @overload
    def __get__(self, instance: Any, _: Type[Any]) -> T_co: ...

    def __get__(self, instance: Any, _: Type[Any]) -> T_co | Self:
        """
        If the descriptor is accessed on an instance, returns the value of
        the descriptor on the given instance.

        If the descriptor is accessed on the owner class, i.e. if
        ``instance`` is :obj:`None`, returns the descriptor object itself.

        :meta public:
        """


class DescriptorBase(TypedDescriptor[T]):
    """
    Base class for descriptors backed by an attribute whose name and access mode
    is determined by the following logic.

    Logic for using ``__dict__`` vs "attr" functions for access to the
    backing attribute:

    1. If the ``use_dict`` argument is set to :obj:`True` in the descriptor
       constructor, then ``__dict__`` will be used. If the library is certain
       that ``__dict__`` is not available on instances of the descriptor owner
       class (cf. :func:`is_dict_available`), then a :obj:`TypeError` is raised
       at the time when ``__set_name__`` is called.
    2. If the ``use_dict`` argument is set to :obj:`False` in the descriptor
       constructor, then the "attr" functions :func:`getattr`, :func:`setattr`,
       :func:`delattr` and :func:`hasattr` will be used. If the library is
       certain that ``__dict__`` is not available on instances of the descriptor
       owner class (cf. :func:`is_dict_available`) and the backing attribute
       name is not present in the class slots (cf. :func:`class_slots`), then a
       :obj:`TypeError` is raised at the time when ``__set_name__`` is called.
    3. If ``use_dict`` is set to :obj:`None` (default value) in the descriptor
       constructor, then :func:`is_dict_available` is called and ``use_dict`` is
       set to the resulting value. Further validation is then performed as
       described in points 1. and 2. above.

    Naming logic for the backing attribute:

    1. If the ``backed_by`` argument is specified in the descriptor constructor,
       the string passed to it is used as name for the backing attribute.
    2. Else, if using ``__dict__`` for access to the backing attribute, then the
       backing attribute name coincides with the descriptor name.
    3. Else, the backing attribute name is obtained by prepending one or
       two underscores to the descriptor name (one if the descriptor name starts
       with underscore, two if it doesn't).

    If the backing attribute name starts with two underscores but does not end
    with two underscores, name-mangling is automatically performed.
    """

    # Attributes set by constructor:
    __type: Union[Type[T], Any]

    # Attributes set by __set_name__:
    __name: str
    __owner: Type[Any]
    __backed_by: str
    __use_dict: bool

    # Attribute set by constructor and deleted by __set_name__:
    __temp_use_dict: Optional[bool]
    __temp_backed_by: Optional[str]

    __slots__ = (
        "__type",
        "__name",
        "__owner",
        "__backed_by",
        "__use_dict",
        "__temp_use_dict",
        "__temp_backed_by",
        "__descriptor_type__",
    )

    @overload
    def __init__(
        self,
        type: Type[T],
        /,
        *,
        backed_by: Optional[str] = None,
        use_dict: Optional[bool] = None,
    ) -> None:
        # pylint: disable = redefined-builtin
        ...

    @overload
    def __init__(
        self,
        type: Any,
        /,
        *,
        backed_by: Optional[str] = None,
        use_dict: Optional[bool] = None,
    ) -> None:
        # pylint: disable = redefined-builtin
        ...

    def __init__(
        self,
        type: Type[T] | Any,  # will be TypeForm[T] one day
        /,
        *,
        backed_by: Optional[str] = None,
        use_dict: Optional[bool] = None,
    ) -> None:
        """
        Creates a new descriptor with the given type and optional validator.

        :param type: the type of the descriptor.
        :param backed_by: name for the backing attribute (optional, default name
                          used if not specified).
        :param use_dict: whether to use ``__dict__`` or slots for access to the
                         backing attribute (optional, automatically determined
                         if not specified).

        :raises TypeError: if the type cannot be validated by the
                           :mod:`typing_validation` library.

        :meta public:
        """
        # pylint: disable = redefined-builtin
        if not can_validate(type):
            raise TypeError(f"Cannot validate type {type!r}.")
        validate(backed_by, Optional[str])
        self.__type = type
        self.__temp_backed_by = backed_by
        self.__temp_use_dict = use_dict
        self.__descriptor_type__ = type

    @final
    @property
    def name(self) -> str:
        """
        The name of the attribute.
        """
        return self.__name

    @final
    @property
    def type(self) -> Union[Type[T], Any]:  # will be TypeForm[T] one day
        """
        The type of the attribute.
        """
        return self.__type

    @final
    @property
    def owner(self) -> Type[Any]:
        """
        The class that owns the attribute.
        """
        return self.__owner

    @final
    def _get_on(self, instance: Any) -> T:
        """
        Gets the value of the backing attribute on the given instance.
        """
        if self.__use_dict:
            try:
                return cast(T, instance.__dict__[self.__backed_by])
            except KeyError:
                raise AttributeError(
                    f"{self.__owner.__name__!r} object has no "
                    f"attribute {self.__backed_by!r}."
                ) from None
        return cast(T, getattr(instance, self.__backed_by))

    @final
    def _set_on(self, instance: Any, value: T) -> None:
        """
        Sets the value of the backing attribute on the given instance.
        """
        if self.__use_dict:
            instance.__dict__[self.__backed_by] = value
        else:
            setattr(instance, self.__backed_by, value)

    @final
    def _del_on(self, instance: Any) -> None:
        """
        Deletes the value of the backing attribute on the given instance.
        """
        if self.__use_dict:
            try:
                del instance.__dict__[self.__backed_by]
            except KeyError:
                raise AttributeError(
                    f"{self.__owner.__name__!r} object has no "
                    f"attribute {self.__backed_by!r}."
                ) from None
        else:
            delattr(instance, self.__backed_by)

    @final
    def _is_set_on(self, instance: Any) -> bool:
        """
        Checkes whether the value of the backing attribute is set on the
        given instance.
        """
        if self.__use_dict:
            return self.__backed_by in instance.__dict__
        return hasattr(instance, self.__backed_by)

    @final
    @property
    def is_assigned(self) -> bool:
        """
        Whether the descriptor has been assigned its owner and name.
        """
        return hasattr(self, "_DescriptorBase__owner")

    @final
    def __set_name__(self, owner: Type[Any], name: str) -> None:
        """
        Hook called when the descriptor is assigned to a class attribute.
        Responsible for:

        - Setting the owner and name of the descriptor
        - Setting the name of the backing attribute (incl. name-mangling)
        - Determining how to access the backing attribute.

        See :class:`DescriptorBase` class documentation for the logic behind
        the backing attribute's name and access mode.

        :raises TypeError: if the descriptor is assigned more than once
        :raises TypeError: if the owner class has ``__slots__`` and the
                           descriptor ``name`` appears in ``__slots__``
        :raises TypeError: if the owner class has ``__slots__``, ``__dict__`` is
                           not in ``__slots__``, and the backing attribute name
                           is not in ``__slots__``.

        :meta public:
        """
        # 1. Ensure descriptor is not assigned twice:
        if self.is_assigned:
            raise TypeError(
                "Cannot set owner/name for the same descriptor twice."
            )
        # 2. If descriptor name is mangled, unmangle it:
        name = name_unmangle(owner, name)
        # 3. Compute backing attribute name and whether to use __dict__:
        temp_backed_by = self.__temp_backed_by
        temp_use_dict = self.__temp_use_dict
        dict_available = is_dict_available(owner)
        owner_slots = class_slots(owner)
        use_dict = dict_available if temp_use_dict is None else temp_use_dict
        if use_dict:
            if not dict_available:
                raise TypeError(
                    "Cannot set use_dict=True in descriptor constructor: "
                    "__dict__ not available on instances of owner class."
                )
            if temp_backed_by is None:
                temp_backed_by = name
        else:
            if owner_slots is None:
                raise TypeError(
                    "Slots are not available on descriptor owner class: "
                    "please set use_dict=True in the descriptor constructor."
                )
            if temp_backed_by is None:
                if name.startswith("_"):
                    temp_backed_by = f"_{name}"
                else:
                    temp_backed_by = f"__{name}"
            if temp_backed_by not in owner_slots:
                raise TypeError(
                    f"Backing attribute {temp_backed_by!r} does not appear in "
                    "the slots of the descriptor owner class. You can either: "
                    "(i) add the backing attribute name to the __slots__, or "
                    "(ii) set use_dict=True in the descriptor constructor, "
                    "as long as __dict__ is available on owner class instances."
                )
        backed_by = name_mangle(owner, temp_backed_by)
        # 4. Set owner, name (not name-mangled) and backed_by (name-mangled):
        self.__owner = owner
        self.__name = name
        self.__backed_by = backed_by
        self.__use_dict = use_dict
        del self.__temp_backed_by

    @abstractmethod
    @overload
    def __get__(self, instance: None, _: Type[Any]) -> Self: ...

    @abstractmethod
    @overload
    def __get__(self, instance: Any, _: Type[Any]) -> T: ...

    @abstractmethod
    def __get__(self, instance: Any, _: Type[Any]) -> T | Self:
        """
        If the descriptor is accessed on an instance, returns the value of
        the descriptor on the given instance.

        If the descriptor is accessed on the owner class, i.e. if
        ``instance`` is :obj:`None`, returns the descriptor object.

        :meta public:
        """



# pylint: disable = missing-docstring

from typing import Any, Optional
import pytest
from typed_descriptors import Attr

@pytest.mark.parametrize("decorator", [False, True])
@pytest.mark.parametrize("x", [0])
@pytest.mark.parametrize("backed_by", [None, "x", "_x", "_y", "__x", "__y"])
def test_single_attr_valid(x: Any, decorator: bool, backed_by: Optional[str]) -> None:
    class C:
        if decorator:
            # pylint: disable = method-hidden
            if backed_by is None:
                @Attr.validator
                def x(self, value: int) -> bool:
                    return value >= 0
            else:
                @Attr.validator(backed_by=backed_by)
                def x(self, value: int) -> bool:
                    return value >= 0
        else:
            x = Attr(int, lambda _, x: x >= 0, backed_by=backed_by)
        def __init__(self, x: int) -> None:
            self.x = x
    C(x)

@pytest.mark.parametrize("decorator", [False, True])
@pytest.mark.parametrize("x", [0])
@pytest.mark.parametrize("backed_by", [None, "_x", "_y", "__x", "__y"])
def test_single_attr_valid_slots(x: Any, decorator: bool, backed_by: Optional[str]) -> None:
    if backed_by is None:
        backing_attr = "__x"
    else:
        backing_attr = backed_by
    class C:
        if decorator:
            # pylint: disable = method-hidden
            if backed_by is None:
                @Attr.validator
                def x(self, value: int) -> bool:
                    return value >= 0
            else:
                @Attr.validator(backed_by=backed_by)
                def x(self, value: int) -> bool:
                    return value >= 0
        else:
            x = Attr(int, lambda _, x: x >= 0, backed_by=backed_by)

        __slots__ = (backing_attr,)
        def __init__(self, x: int) -> None:
            self.x = x
    C(x)


@pytest.mark.parametrize("decorator", [False, True])
@pytest.mark.parametrize("x", [1])
@pytest.mark.parametrize("backed_by", [None, "x", "_x", "_y", "__x", "__y"])
def test_single_attr_readonly(x: Any, decorator: bool, backed_by: Optional[str]) -> None:
    class C:
        if decorator:
            # pylint: disable = method-hidden
            @Attr.validator(readonly=True, backed_by=backed_by)
            def x(self, value: int) -> bool:
                return value >= 0
        else:
            x = Attr(int, lambda _, x: x >= 0, readonly=True, backed_by=backed_by)
        def __init__(self, x: int) -> None:
            self.x = x
    C(x)
    with pytest.raises(AttributeError):
        C(0).x = x

@pytest.mark.parametrize("decorator", [False, True])
@pytest.mark.parametrize("x", [1])
@pytest.mark.parametrize("backed_by", [None, "_x", "_y", "__x", "__y"])
def test_single_attr_readonly_slots(x: Any, decorator: bool, backed_by: Optional[str]) -> None:
    if backed_by is None:
        backing_attr = "__x"
    else:
        backing_attr = backed_by
    class C:
        if decorator:
            # pylint: disable = method-hidden
            @Attr.validator(readonly=True, backed_by=backed_by)
            def x(self, value: int) -> bool:
                return value >= 0
        else:
            x = Attr(int, lambda _, x: x >= 0, readonly=True, backed_by=backed_by)
        __slots__ = (backing_attr,)
        def __init__(self, x: int) -> None:
            self.x = x
    C(x)
    with pytest.raises(AttributeError):
        C(0).x = x

@pytest.mark.parametrize("decorator", [False, True])
@pytest.mark.parametrize("x, y", [(0, 2), (1, 1)])
def test_two_attr_valid(x: Any, y: Any, decorator: bool) -> None:
    class C:
        if decorator:
            # pylint: disable = method-hidden
            @Attr.validator
            def x(self, value: int) -> bool:
                return value >= 0
            @Attr.validator
            def y(self, value: int) -> bool:
                return value >= self.x
        else:
            x = Attr(int, lambda _, x: x >= 0)
            y = Attr(int, lambda self, y: y >= self.x)
        def __init__(self, x: int, y: int) -> None:
            self.x = x
            self.y = y
    C(x, y)

@pytest.mark.parametrize("decorator", [False, True])
@pytest.mark.parametrize("x, y", [(0, 2), (1, 1)])
def test_two_attr_valid_slots(x: Any, y: Any, decorator: bool) -> None:
    class C:
        if decorator:
            # pylint: disable = method-hidden
            @Attr.validator
            def x(self, value: int) -> bool:
                return value >= 0
            @Attr.validator
            def y(self, value: int) -> bool:
                return value >= self.x
        else:
            x = Attr(int, lambda _, x: x >= 0)
            y = Attr(int, lambda self, y: y >= self.x)
        __slots__ = ("__x", "__y")
        def __init__(self, x: int, y: int) -> None:
            self.x = x
            self.y = y
    C(x, y)

@pytest.mark.parametrize("decorator", [False, True])
@pytest.mark.parametrize("x", ["hello", 1.0])
def test_single_attr_type_error(x: Any, decorator: bool) -> None:
    class C:
        if decorator:
            # pylint: disable = method-hidden
            @Attr.validator
            def x(self, value: int) -> bool:
                return value >= 0
        else:
            x = Attr(int, lambda _, x: x >= 0)
        def __init__(self, x: int) -> None:
            self.x = x
    with pytest.raises(TypeError):
        C(x)
    with pytest.raises(TypeError):
        c = C(0)
        c.x = x

@pytest.mark.parametrize("decorator", [False, True])
@pytest.mark.parametrize("x", [-1])
def test_single_attr_value_error(x: Any, decorator: bool) -> None:
    class C:
        if decorator:
            # pylint: disable = method-hidden
            @Attr.validator
            def x(self, value: int) -> bool:
                return value >= 0
        else:
            x = Attr(int, lambda _, x: x >= 0)
        def __init__(self, x: int) -> None:
            self.x = x
    with pytest.raises(ValueError):
        C(x)
    with pytest.raises(ValueError):
        C(0).x = x

@pytest.mark.parametrize("decorator", [False, True])
@pytest.mark.parametrize("x, y", [(0, -1), (-1, 2), (3, 2)])
def test_two_attr_value_error(x: Any, y: Any, decorator: bool) -> None:
    class C:
        if decorator:
            # pylint: disable = method-hidden
            @Attr.validator
            def x(self, value: int) -> bool:
                return value >= 0
            @Attr.validator
            def y(self, value: int) -> bool:
                return value >= self.x
        else:
            x = Attr(int, lambda _, x: x >= 0)
            y = Attr(int, lambda self, y: y >= self.x)
        def __init__(self, x: int, y: int) -> None:
            self.x = x
            self.y = y
    with pytest.raises(ValueError):
        C(x, y)
    with pytest.raises(ValueError):
        C(x, x+1).y = y



r"""
    Autodoc extension dealing with local type references and function signatures.
"""

# Some References:
#
# https://www.sphinx-doc.org/en/master/development/tutorials/autodoc_ext.html
# https://github.com/sphinx-doc/sphinx/blob/49d1e7142eca6d303f1e929bdbe9d8e4749858b7/sphinx/ext/autodoc/__init__.py#L2737
# https://stackoverflow.com/questions/39565203/how-to-specialize-documenters-in-sphinx-autodoc


from __future__ import annotations

from collections import deque
from collections.abc import Callable, Iterable, Iterator, Mapping
from dataclasses import dataclass
import inspect
import re
from types import FunctionType, ModuleType
from typing import Any, Optional, TypeVar
from sphinx.application import Sphinx
from typed_descriptors import TypedDescriptor
from typing_validation import inspect_type


### 1. Parse Type Annotations ###

@dataclass(frozen=True)
class ParsedType:
    r""" Dataclass for a parsed type. """
    name: str
    args: None|str|tuple[ParsedType, ...] = None
    variadic: bool = False

    def __post_init__(self) -> None:
        name, args, variadic = self.name, self.args, self.variadic
        assert isinstance(name, str)
        assert isinstance(variadic, bool)
        if args is not None and not isinstance(args, str):
            assert isinstance(args, tuple)
            assert all(isinstance(arg, ParsedType) for arg in args)
        if variadic:
            assert isinstance(args, tuple)
        if "(" in name or ")" in name:
            raise ValueError(
                "Round brackets in type annotations are only supported for "
                "empty type arguments lists, in the form TypeName[()]."
            )
        if isinstance(args, str) and not args:
            raise ValueError(
                "Literal type must include at least one value."
            )

    def crossref(self, globalns: Optional[Mapping[str, Any]] = None) -> str:
        r""" Generates Sphinx cross-reference link for the given type, using local names. """
        # pylint: disable = eval-used
        if globalns is None:
            globalns = {}
        name, args, variadic = self.name, self.args, self.variadic
        role = "obj"
        if name in globalns:
            obj = globalns[name]
            if isinstance(obj, ModuleType):
                role = "mod"
            elif isinstance(obj, property):
                role = "attr"
            elif isinstance(obj, type):
                if name not in ("Any", "typing.Any"):
                    role = "class"
            elif isinstance(obj, FunctionType):
                role = "func"
        name_crossref = f":{role}:`{name}`"
        if args is None:
            return name_crossref
        if name in ("UnionType", "types.UnionType"):
            assert isinstance(args, tuple)
            return " | ".join((arg.crossref(globalns) for arg in args))
        if isinstance(args, str):
            _args = eval(f"({args}, )")
            arg_crossrefs = ", ".join(f"``{repr(arg)}``" for arg in _args)
        elif not args:
            arg_crossrefs = "()"
        else:
            arg_crossrefs = ", ".join((arg.crossref(globalns) for arg in args))
        if variadic:
            arg_crossrefs += ", ..."
        return fr"{name_crossref}\ [{arg_crossrefs}]"

    def _repr(self, level: int = 0) -> list[str]:
        basic_indent = "  "
        indent = basic_indent * level
        next_indent = basic_indent * (level+1)
        name, args, variadic = self.name, self.args, self.variadic
        if args is None:
            assert not variadic
            return [indent+f"ParsedType({name!r})"]
        lines = [
            indent+"ParsedType(",
            next_indent+f"name = {self.name!r},"
        ]
        if isinstance(args, str):
            assert not variadic
            lines.append(next_indent+f"args = {args!r}")
        else:
            assert isinstance(args, tuple)
            if not args:
                assert not variadic
                lines.append(next_indent+"args = ()")
            else:
                lines.append(next_indent+"args = (")
                for i, arg in enumerate(args):
                    assert isinstance(arg, ParsedType)
                    lines.extend(arg._repr(level+2))
                    sep = "," if i < len(args)-1 or len(args) == 1 else ""
                    lines[-1] = lines[-1]+sep
                lines.append(next_indent+")")
        if variadic:
            lines[-1] = lines[-1]+","
            lines.append(next_indent+f"variadic = True")
        lines.append(indent+")")
        return lines

    def __repr__(self) -> str:
        """
        Structured representation of the parsed type.
        """
        return "\n".join(self._repr())

def _outer_bracket_ranges(
    s: str,
    start: int, stop: int
) -> Iterator[range]:
    if stop is None:
        stop = len(s)
    open: int|None = None
    level = 0
    for i in range(start, stop):
        c = s[i]
        if c == "[":
            if open is None:
                assert level == 0
                open = i
            level += 1
        if c == "]":
            if open is None:
                raise ValueError(f"Unbalanced ']' at index {i}.")
            assert level > 0
            level -= 1
            if level == 0:
                yield range(open, i+1)
                open = None
    if open is not None:
        raise ValueError(f"Unbalanced '[' at index {open}.")

def _find_outside_ranges(
    char: str, s: str, ranges: Iterable[range],
    start: int, stop: int
) -> Iterator[int]:
    assert len(char) == 1, f"Expected single char, found {s!r}."
    if stop is None:
        stop = len(s)
    ranges = deque(sorted(ranges, key=lambda r: r.start))
    _start = start
    while (char_idx := s.find(char, _start, stop)) >= 0:
        while ranges and ranges[0].stop <= _start:
            ranges.popleft()
        if not ranges or char_idx not in ranges[0]:
            yield char_idx
            _start = char_idx+1
        else:
            r = ranges.popleft()
            _start = r.stop

def _split_at(
    idxs: Iterable[int],
    start: int, stop: int
) -> Iterable[range]:
    idxs = sorted(idxs)
    assert all(idx in range(start, stop) for idx in idxs)
    _start = start
    for idx in idxs:
        if idx > _start:
            yield range(_start, idx)
        else:
            assert idx == _start
        _start = idx + 1
    if _start < stop:
        yield range(_start, stop)

def _parsed_type(
    annotation: str,
    start: int,
    stop: int,
    name: str, args:
    None|str|tuple[ParsedType, ...] = None,
    variadic: bool = False
) -> ParsedType:
    try:
        t = ParsedType(name, args, variadic)
    except ValueError as e:
        raise ValueError(
            "Error parsing type at \n"
            f"{start = }, {stop = }, {annotation[start:stop] = }.\n"
            f"ValueError: {e}"
        )
    return t

def _parse_type_args(
    annotation: str, start: int, stop: int
) -> tuple[tuple[ParsedType, ...], bool]:
    # print(f"_parse_type_args({annotation!r}, {start}, {stop})")
    # print(f"{annotation[start:stop] = }")
    if annotation[start:stop].strip() == "()":
        return (), False
    bracket_ranges = tuple(_outer_bracket_ranges(annotation, start, stop))
    comma_idxs = tuple(_find_outside_ranges(",", annotation, bracket_ranges, start, stop))
    if not comma_idxs:
        return (_parse_type(annotation, start, stop),), False
    arg_ranges = tuple(_split_at(comma_idxs, start, stop))
    args: list[ParsedType] = []
    variadic = False
    for i, r in enumerate(arg_ranges):
        arg = _parse_type(annotation, r.start, r.stop)
        if arg.name == "...":
            if i < len(arg_ranges)-1:
                raise ValueError(
                    "Ellipsis found in args, but not in last position,  at "
                    f"{start = }, {stop = }, {annotation[start:stop] = }"
                )
            variadic = True
            break
        args.append(arg)
    return tuple(args), variadic

def _parse_atom_type(annotation: str, start: int, stop: int) -> ParsedType:
    # print(f"_parse_atom_type({annotation!r}, {start}, {stop})")
    # print(f"{annotation[start:stop] = }")
    while start < stop and annotation[start].isspace():
        start += 1
    while start < stop and annotation[stop-1].isspace():
            stop -= 1
    assert annotation[start:stop] == annotation[start:stop].strip()
    bracket_ranges = tuple(_outer_bracket_ranges(annotation, start, stop))
    if len(bracket_ranges) > 1:
        raise ValueError(
            "Non-union type must take the form 'TypeName' or 'TypeName[Args]'. "
            f"Found multiple outer bracket pairs at "
            f"{start = }, {stop = }, {annotation[start:stop] = }"
        )
    r = bracket_ranges[0] if bracket_ranges else None
    if r is None:
        name = annotation[start:stop].strip()
        return _parsed_type(annotation, start, stop, name)
    elif r.stop < stop:
        raise ValueError(
            "Non-union type must take the form 'TypeName' or 'TypeName[Args]'. "
            "Found text after bracket pair at "
            f"{start = }, {stop = }, {annotation[start:stop] = }"
        )
    name = annotation[start:r.start].strip()
    if name in ("Literal", "typing.Literal"):
        return _parsed_type(annotation, start, stop, name, annotation[r.start+1:r.stop-1])
    if not name:
        raise ValueError(
            f"Found empty type name at "
            f"{start = }, {stop = }, {annotation[start:stop] = }"
        )
    args, variadic = _parse_type_args(annotation, r.start+1, r.stop-1)
    return _parsed_type(annotation, start, stop, name, args, variadic)

def _parse_type(annotation: str, start: int, stop: int) -> ParsedType:
    # print(f"_parse_type({annotation!r}, {start}, {stop})")
    # print(f"{annotation[start:stop] = }")
    bracket_ranges = tuple(_outer_bracket_ranges(annotation, start, stop))
    ors_idxs = tuple(_find_outside_ranges("|", annotation, bracket_ranges, start, stop))
    if not ors_idxs:
        return _parse_atom_type(annotation, start, stop)
    member_ranges = tuple(_split_at(ors_idxs, start, stop))
    name = "UnionType"
    args = tuple(
        _parse_atom_type(annotation, r.start, r.stop)
        for r in member_ranges
    )
    return _parsed_type(annotation, start, stop, name, args)

def parse_type(annotation: str) -> ParsedType:
    return _parse_type(annotation, 0, len(annotation))


### 2. Track Classes Known to Autodoc ###

_class_dict: dict[str, type] = {}

def class_tracking_handler(app: Sphinx, what: str, fullname: str, obj: Any, options: Any, lines: list[str]) -> None:
    r"""
        Handler for Sphinx Autodoc's event
        `autodoc-process-docstring <https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html#event-autodoc-process-docstring>`_
        which keeps track of classes known to autodoc, for used in other handlers.
    """
    if what != "class":
        return
    _class_dict[fullname] = obj


### 3. Document Function Parameter and Return Types ###

def _sigdoc(fun: FunctionType, lines: list[str]) -> None:
    r"""
        Returns doclines documenting the parameter and return type of the given function
    """
    # pylint: disable = too-many-branches
    doc = "\n".join(lines)
    lines.append("")
    # FIXME: if an :rtype: line already exists, remove it here and re-append it after all param type lines.
    globalns = fun.__globals__
    sig = inspect.signature(fun)
    for p in sig.parameters.values():
        annotation = p.annotation
        if annotation == p.empty:
            continue
        if not isinstance(annotation, str):
            print(f"WARNING! Found non-string annotation: {repr(annotation)}. Did you forget to import annotation from __future__?.")
            annotation = str(annotation)
        t = parse_type(annotation)
        tx = t.crossref(globalns)
        default = p.default if p.default != p.empty else None
        is_args = p.kind == p.VAR_POSITIONAL
        is_kwargs = p.kind == p.VAR_KEYWORD
        if is_args:
            extra_info = "variadic positional"
        elif is_kwargs:
            extra_info = "variadic keyword"
        elif default is not None:
            default_str = default.__qualname__ if isinstance(default, FunctionType) else repr(default)
            extra_info = f"default = ``{default_str}``"
        else:
            extra_info = None
        if extra_info is None:
            line = f":type {p.name}: {tx}"
        else:
            line = f":type {p.name}: {tx}; {extra_info}"
        if f":param {p.name}:" not in doc:
            lines.append(f":param {p.name}:")
        if f":type {p.name}:" not in doc:
            lines.append(line)
    if sig.return_annotation == sig.empty:
        return
    t = parse_type(sig.return_annotation)
    tx = t.crossref(globalns)
    line = f":rtype: {tx}"
    if ":rtype:" not in doc:
        lines.append(line)

def signature_doc_handler(app: Sphinx, what: str, fullname: str, obj: Any, options: Any, lines: list[str]) -> None:
    r"""
        Handler for Sphinx Autodoc's event
        `autodoc-process-docstring <https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html#event-autodoc-process-docstring>`_
        which automatically documents parameter types and return type for
        functions, methods and properties.
    """
    # pylint: disable = too-many-arguments
    if what not in ("function", "method", "property"):
        return
    if what == "property":
        fun: FunctionType = obj.fget
    else:
        fun = obj
    _sigdoc(fun, lines)


### 4. Document Attribute Types ###

def attr_doc_handler(app: Sphinx, what: str, fullname: str, obj: Any, options: Any, lines: list[str]) -> None:
    r"""
        Handler for Sphinx Autodoc's event
        `autodoc-process-docstring <https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html#event-autodoc-process-docstring>`_
        which automatically documents the type of attributes, including those
        defined by typed descriptors.
    """
    # pylint: disable = too-many-arguments
    if what != "attribute":
        return
    attrname = fullname.split(".")[-1]
    classname = ".".join(fullname.split(".")[:-1])
    parent_class = _class_dict.get(classname)
    if parent_class is not None:
        annotations = parent_class.__annotations__
        if attrname in annotations:
            type_annotation = annotations[attrname]
        elif isinstance(obj, TypedDescriptor):
            ty = obj.__descriptor_type__
            inspector = inspect_type(ty)
            type_annotation = inspector.type_annotation
        else:
            type_annotation = None
        if type_annotation is not None:
            t = parse_type(type_annotation)
            tx = t.crossref()
            if ":rtype:" not in "\n".join("lines"):
                lines.append(f":rtype: {tx}")


### 5. Replace Cross-references with Fully Qualified Names ###

def simple_crossref_pattern(name: str) -> re.Pattern[str]:
    r"""
        Pattern for simple imports:

        .. code-block :: python

            f":{role}:`{name}`"        # e.g. ":class:`MyClass`"
            f":{role}:`~{name}`"       # e.g. ":class:`~MyClass`"
            f":{role}:`{name}{tail}`"  # e.g. ":attr:`MyClass.my_property.my_subproperty`"
            f":{role}:`~{name}{tail}`" # e.g. ":attr:`~MyClass.my_property.my_subproperty`"

    """
    return re.compile(rf":([a-z]+):`(~)?{name}(\.[\.a-zA-Z0-9_]+)?`")

def simple_crossref_repl(name: str, fullname: str) -> Callable[[re.Match[str]], str]:
    r"""
        Replacement function for the pattern generated by :func:`simple_crossref_pattern`:

        .. code-block :: python

            f":{role}:`~{fullname}`"                    # e.g. ":class:`~mymod.mysubmod.MyClass`"
            f":{role}:`~{fullname}`"                    # e.g. ":class:`~mymod.mysubmod.MyClass`"
            f":{role}:`{name}{tail}<{fullname}{tail}>`" # e.g. ":attr:`MyClass.my_property.my_subproperty<mymod.mysubmod.MyClass.my_property.my_subproperty>`"
            f":{role}:`~{fullname}{tail}`"              # e.g. ":attr:`~mymod.mysubmod.MyClass.my_property.my_subproperty`"

    """
    def repl(match: re.Match[str]) -> str:
        role = match[1]
        short = match[2] is not None
        tail = match[3]
        if tail is None:
            return f":{role}:`~{fullname}`"
        if short:
            return f":{role}:`~{fullname}{tail}`"
        return f":{role}:`{name}{tail}<{fullname}{tail}>`"
    return repl

def labelled_crossref_pattern(name: str) -> re.Pattern[str]:
    r"""
        Pattern for labelled imports:

        .. code-block :: python

            f":{role}:`{label}<{name}>`"       # e.g. ":class:`my class<MyClass>`"
            f":{role}:`{label}<{name}{tail}>`" # e.g. ":attr:`my_property<MyClass.my_property>`"

    """
    return re.compile(rf":([a-z]+):`([\.a-zA-Z0-9_]+)<{name}(\.[\.a-zA-Z0-9_]+)?>`")

def labelled_crossref_repl(name: str, fullname: str) -> Callable[[re.Match[str]], str]:
    r"""
        Replacement function for the pattern generated by :func:`labelled_crossref_pattern`:

        .. code-block :: python

            f":{role}:`{label}<{fullname}>`"       # e.g. ":class:`my class<mymod.mysubmod.MyClass>`"
            f":{role}:`{label}<{fullname}{tail}>`" # e.g. ":attr:`my_property<mymod.mysubmod.MyClass.my_property>`"

    """
    def repl(match: re.Match[str]) -> str:
        role = match[1]
        label = match[2]
        tail = match[3]
        if tail is None:
            return f":{role}:`{label}<{fullname}>`"
        return f":{role}:`{label}<{fullname}{tail}>`"
    return repl

_crossref_subs: list[tuple[Callable[[str], re.Pattern[str]],
                           Callable[[str, str], Callable[[re.Match[str]], str]]]] = [
    (simple_crossref_pattern, simple_crossref_repl),
    (labelled_crossref_pattern, labelled_crossref_repl),
]
r"""
    Substitution patterns and replacement functions for various kinds of cross-reference scenarios.
"""

def _get_module_by_name(modname: str) -> ModuleType:
    r"""
        Gathers a module object by name.
    """
    # pylint: disable = exec-used, eval-used
    exec(f"import {modname.split('.')[0]}")
    mod: ModuleType = eval(modname)
    if not isinstance(mod, ModuleType):
        return None
    return mod

def _get_obj_mod(app: Sphinx, what: str, fullname: str, obj: Any) -> Optional[ModuleType]:
    r"""
        Gathers the containing module for the given ``obj``.
    """
    autodoc_type_aliases = app.config.__dict__.get("autodoc_type_aliases")
    name = fullname.split(".")[-1]
    obj_mod: Optional[ModuleType]
    if autodoc_type_aliases is not None:
        if name in autodoc_type_aliases and fullname == autodoc_type_aliases[name]:
            modname = ".".join(fullname.split(".")[:-1])
            obj_mod = _get_module_by_name(modname)
            return obj_mod
    if what == "module":
        obj_mod = obj
    elif what in ("function", "class", "method", "exception"):
        obj_mod = inspect.getmodule(obj)
    elif what == "property":
        obj_mod = inspect.getmodule(obj.fget)
    elif what == "data":
        modname = ".".join(fullname.split(".")[:-1])
        obj_mod = _get_module_by_name(modname)
    elif what == "attribute":
        modname = ".".join(fullname.split(".")[:-2])
        obj_mod = _get_module_by_name(modname)
    else:
        print(f"WARNING! Encountered unexpected value for what = {what} at fullname = {fullname}")
        obj_mod = None
    return obj_mod

def _build_fullname_dict(app: Sphinx, fullname: str, obj_mod: Optional[ModuleType], ) -> dict[str, str]:
    r"""
        Builds a dictionary of substitutions from module global names to their fully qualified names,
        based on :func:`inspect.getmodule` and `autodoc_type_aliases` (if specified in the Sphinx app config).
    """
    autodoc_type_aliases = app.config.__dict__.get("autodoc_type_aliases")
    fullname_dict: dict[str, str] = {}
    if obj_mod is not None:
        globalns = obj_mod.__dict__
        for g_name, g_obj in globalns.items():
            if isinstance(g_obj, (FunctionType, type)):
                g_mod = inspect.getmodule(g_obj)
            elif isinstance(g_obj, ModuleType):
                g_mod = g_obj
            else:
                g_mod = inspect.getmodule(g_obj)
            if g_mod is None or g_mod == obj_mod:
                continue
            if g_name not in g_mod.__dict__:
                continue
            g_modname = g_mod.__name__
            fullname_dict[g_name] = f"{g_modname}.{g_name}"
    if autodoc_type_aliases is not None:
        for a_name, a_fullname in autodoc_type_aliases.items():
            if a_name not in fullname_dict:
                fullname_dict[a_name] = a_fullname
    return fullname_dict

def local_crossref_handler(app: Sphinx, what: str, fullname: str, obj: Any, options: Any, lines: list[str]) -> None:
    r"""
        Handler for Sphinx Autodoc's event
        `autodoc-process-docstring <https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html#event-autodoc-process-docstring>`_
        which replaces cross-references specified in terms of module globals with their fully qualified version.
    """
    # pylint: disable = too-many-arguments, too-many-locals
    obj_mod = _get_obj_mod(app, what, fullname, obj)
    fullname_dict = _build_fullname_dict(app, fullname, obj_mod)
    for sub_name, sub_fullname in fullname_dict.items():
        for idx, line in enumerate(lines):
            for pattern_fun, repl_fun in _crossref_subs:
                pattern = pattern_fun(sub_name)
                repl = repl_fun(sub_name, sub_fullname)
                line = re.sub(pattern, repl, line)
            lines[idx] = line


### 6. Register Sphinx Event Handlers ###

def setup(app: Sphinx) -> None:
    r"""
        Registers handlers for Sphinx events.
    """
    app.connect("autodoc-process-docstring", class_tracking_handler)
    app.connect("autodoc-process-docstring", signature_doc_handler)
    app.connect("autodoc-process-docstring", attr_doc_handler)
    app.connect("autodoc-process-docstring", local_crossref_handler)
